/*************************************************************************
	-----------------------------------------------------
           		Copyright 2002, 
	   University of Massachusetts, Amherst. 
	-----------------------------------------------------

This package was created at the University of Massachusetts, Amherst. 
The University of Massachusetts makes no warranty about the suitability 
of this software for any purpose. It is presented on an "as is" basis.

CONDITIONS:

The use of software for commercial purposes is strictly forbidden
without a written consent from its authors or the University of 
Massachusetts, Amherst.

**************************************************************************/
/**CHeaderFile*****************************************************************

  FileName    [build.h]

  PackageName [BDS]

  Synopsis    [BDD decomposition-based logic optimization program]

  Description []

  SeeAlso     []

  Author      [Congguang Yang]

  Last date   [1/11/99]

******************************************************************************/

#ifndef _BDSBUILD
#define _BDSBUILD

/*---------------------------------------------------------------------------*/
/* Nested includes                                                           */
/*---------------------------------------------------------------------------*/

#include "bnet.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

#define BDS_GLOBAL_BDD_LIMIT 20000 /* During the process of constructing global
				       ** BDDs, if total number of BDD nodes is larger
				       ** then this number, the construction is aborted
				       */
#define BDS_LOCAL_GLOBAL 8   /* This is pure empirical value. When the size of
				 ** local BDDs is this times larger than global
				 ** BDDs, global BDDs are used for synthesis. By setting
				 ** this value to a even larger one will bias towards
				 ** local BDDs 
				 */
#define PI_PS_FROM_FILE	0
#define PI_PS_DFS	1
#define PI_PS_GIVEN	2

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

typedef	struct	bdsOptions {
    /* Options related to BDS. 10/20/98 */
    long	initialTime;	/* this is here for convenience */
    char	*file;		/* network file name */
    int		verb;		/* level of verbosity */
    int		cacheSize;	/* computed table initial size */
    unsigned long maxMemory;	/* computed table maximum size */
    int		slots;		/* unique subtable initial slots */
    int		maxGrowth;	/* maximum growth during reordering (%) */
    Cudd_ReorderingType	reordering; /* NONE RANDOM PIVOT SIFTING ... */
    int		order; 	        /* FANIN DFS ... */
    char	*orderfile;	/* file for externally provided order */

    /* Options created new for BDS 3/22/98 */
    int		useloc;		/* Use local BDD only */
    int		useglb;		/* Use global BDD only */
    int		autodyn;	/* Enable dynamic variable reordering */
    int		rodelim;	/* Use variable reordering in eliminate */
    int		globalLimit;    /* The limit for BDD nodes number, if > this limit, no global BDDs */
    int		ethred;		/* Elimination gain threshold. */
    int		glbEliminate;   /* Eliminate nodes using global BDDs */
    int		verify;		/* Verify the synthesis results */
    int		effort;		/* Amount of effort users want to put on the process */
    int		limitSize;	/* The #Node * #Var of a BDD, which can apply exact reorder. */
    int		miniMethod;     /* Methods for BDD minimizations. */
    int 	dumpbdd;	/* Dump the BDD of the circuit in dot format */
    int		dumpblif;	/* Dump the ftree into BLIF format */
    int		dumpftree;	/* Dump ftree in the form of linking pointers */
    int		dumpfdot;	/* Dump ftree in dot format */
    int		dumpnet;	/* Dump bnet after eliminate */
    int		cutptl;		/* Cut PTLs from synthesis results */

    long    timeout;    /* a guard against cycling decomposition */
    int     xdelta;     /* reject x-decomp that raises BDD size this much JS */
    char    *dir;       /* forced directory name JS */
    unsigned methods;   /* switch any decomp methods on/off JS */
} bdsOptions;
/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/
#define BDS_NO_TIMEOUT -1

/* These are potential duplicates. */
#ifndef EXTERN
#   ifdef __cplusplus
#	define EXTERN extern "C"
#   else
#	define EXTERN extern
#   endif
#endif
#ifndef ARGS
#   if defined(__STDC__) || defined(__cplusplus)
#	define ARGS(protos)	protos		/* ANSI C */
#   else /* !(__STDC__ || __cplusplus) */
#	define ARGS(protos)	()		/* K&R C */
#   endif /* !(__STDC__ || __cplusplus) */
#endif

#ifndef TRUE
#   define TRUE 1
#endif
#ifndef FALSE
#   define FALSE 0
#endif

/**Macro***********************************************************************

  Synopsis     [Returns 1 if the two arguments are identical strings.]

  Description  []

  SideEffects  [none]

  SeeAlso      []
		   
******************************************************************************/
#define STRING_EQUAL(s1,s2) (strcmp((s1),(s2)) == 0)

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

EXTERN int BDS_BuildGlobalBDD ARGS((DdManager*,BnetNetwork*,bdsOptions*,int*));
EXTERN int BDS_BuildLocalBDD ARGS((DdManager*,BnetNetwork*,bdsOptions*));

EXTERN int BDS_buildDDs ARGS((BnetNetwork*,DdManager*,BnetNode*,bdsOptions*));
EXTERN int BDS_BuildNodeBDD ARGS((DdManager*,BnetNode*,st_table*));
EXTERN void BDS_BDD2ADD ARGS((DdManager*,BnetNode*));

#endif /* _BDSBUILD */
