/*************************************************************************
	-----------------------------------------------------
           		Copyright 2002, 
	   University of Massachusetts, Amherst. 
	-----------------------------------------------------

This package was created at the University of Massachusetts, Amherst. 
The University of Massachusetts makes no warranty about the suitability 
of this software for any purpose. It is presented on an "as is" basis.

CONDITIONS:

The use of software for commercial purposes is strictly forbidden
without a written consent from its authors or the University of 
Massachusetts, Amherst.

**************************************************************************/
/**CFile***********************************************************************

  FileName    [lopt.c]

  PackageName [BDS]

  Synopsis    [BDD Decomposition-Based Logic Optimization Program]

  Description []

  SeeAlso     []

  Author      [Anda Congguang Yang]

  Copyright   [The file was created in Department of Electrical & Computer
	Engineering, University of Massachusetts Amherst. All source codes
	package are CONFIDENTIAL before a formal release is
	announced. The author reserve all right of the package.]

******************************************************************************/

#include "lopt.h"
#include <stdio.h>

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static int BDS_CheckTimeout ARGS((bdsOptions*));
static int BDS_CheckTerminal ARGS((DdManager*,DdNode*,FactorTreeNode*));
static int BDS_FindBestDecomp ARGS((DdManager*,DdNode*,DdNode*,int*,bdsOptions*,bddPool**,bddPool**,int*,int,edgeMark**));
static int BDS_XHardCore ARGS((DdManager*,DdNode*,bdsOptions*,bddPool**,bddPool**,int*,int*,int,edgeMark**));
static int BDS_BHardCore ARGS((DdManager*,DdNode*,DdNode*,bdsOptions*,bddPool**,bddPool**,bddPool**,int*,int*,int,edgeMark**));
static int BDS_XDecompOnBdd ARGS((DdManager*,DdNode*,int,int*,int,edgeMark*,XdecompCost*,bdsOptions*));
static DdNode * BDS_addXCandidate ARGS((DdManager*,edgeMark*));
static DdNode * BDS_bddXCandidate ARGS((DdManager*,DdNode*,edgeMark*));
static int BDS_BddDecomposeMain ARGS((DdManager*,DdNode*,bdsOptions*,bddPool**,bddPool**,bddPool**,int*,int*)); 
static decompCost* BDS_CalculateDecompCost ARGS((DdManager *,DdNode *,DdNode *,DdNode *,bdsOptions *));
static int BDS_PerformDecomp ARGS((DdManager *,DdNode *,DdNode *,edgeMark *,decompCost **,bdsOptions *,int,int));
static int BDS_DDDecomposeAll ARGS((DdManager*,bddPool*,bdsOptions*,FactorTreeNode**)); 
static int BDS_SimpleDominator ARGS((DdManager*,DdNode*,DdNode*,int*,int,int*,edgeMark**,bddPool**,bddPool**,int*,bdsOptions*));
static DdNode * BDS_SpecialCheck ARGS((DdManager*,DdNode*,DdNode*,int*,int,int*,edgeMark**,int*,bdsOptions*));
static int BDS_CofactorOnTop ARGS((DdManager*,DdNode*,bdsOptions*,bddPool**,bddPool**,bddPool**,int*,int*));
static DdNode * BDS_SearchXD ARGS((DdManager*,DdNode*,int*));
static DdNode * BDS_SearchOneD ARGS((DdManager*,int*,int,int*,edgeMark**));
static DdNode * BDS_SearchZeroD ARGS((DdManager*,int*,int,int*,edgeMark**));
static DdNode * BDS_XDecompose ARGS((DdManager*,DdNode*,DdNode*,int*,bdsOptions*));
static DdNode * BDS_ANDDecompose ARGS((DdManager*,DdNode*,DdNode*,bdsOptions*)); 
static DdNode * BDS_ORDecompose ARGS((DdManager*,DdNode*,DdNode*,bdsOptions*)); 
static DdNode * BDS_BranchCandidateX ARGS((DdManager*,DdNode*,edgeMark*));
static DdNode * BDS_BranchCandidateB ARGS((DdManager*,DdNode*,edgeMark*,DdNode**,DdNode**));
static DdNode * BDS_FindOneMark ARGS((DdManager*,DdNode*,edgeMark*));
static DdNode * BDS_FindOneMarkRecursive ARGS((DdManager*,DdNode*,st_table*,edgeMark*));
static int BDS_OneMoreTry ARGS((DdManager*,DdNode*,bdsOptions*,bddPool**,bddPool**,int*,int*));
static decompCost* BDS_InitBestCost ARGS(());
static XdecompCost* BDS_InitXdecompCost ARGS(());
static float BDS_CalXdecompCost ARGS((DdManager*,DdNode*,int,DdNode*,int,DdNode*,int,DdNode*,int));

/**Function********************************************************************
    return TRUE if timeout specified and time elapsed
*/
int BDS_CheckTimeout (bdsOptions* options)
{
    if (options->timeout == BDS_NO_TIMEOUT)
        return FALSE;
    return (util_cpu_time() - options->initialTime > options->timeout);
}
/**Function********************************************************************

  Synopsis    [Main Program for BDD Decomposition-Based BDS]

  Description [result = 1, successful. 0 otherwise]

  SideEffects []

  SeeAlso     []

  LastDate    [3/25/99]

******************************************************************************/
extern
int
BDS_DDDecompose(
  DdManager *bddmgr,
  DdNode *bdd,
  bdsOptions *option,
  FactorTreeNode **FTNode)
{
  int result, operator, sl;
  bddPool *decomposed = NULL, *dominator = NULL,*control = NULL;
  FactorTreeNode *decomposedFTNode, *dominatorFTNode, *controlFTNode;
  int p; /* If decompBdd is complementary or not after reordering */

  if (bdd == NULL) { 
      printf("WARNING: Can not decompose NULL BDD !\n");
      return(0);
  }

#ifndef DEBUG
  (void) printf(".");
  (void) fflush(stdout);
#endif

#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_DDDecompose", 0);
#endif

  /* When the BDD is a terminal, no decomposition */
  result = BDS_CheckTerminal(bddmgr,bdd,*FTNode);
  if (result != 0) return(1);

#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_DDDecompose", 1);
#endif

  /* Decompose the BDD */
  result = BDS_BddDecomposeMain(bddmgr,bdd,option,&decomposed,&dominator,&control,&operator,&p);
  if (result == 0) { return 0;}

#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_DDDecompose", 2);
#endif

  /* Build factoring tree */
  decomposedFTNode = ALLOC(FactorTreeNode, 1); 
  DBG_MEMSET (decomposedFTNode, 0, sizeof (FactorTreeNode));
  (*FTNode)->siblings[0] = decomposedFTNode;
  dominatorFTNode = ALLOC(FactorTreeNode, 1); 
  DBG_MEMSET (dominatorFTNode, 0, sizeof (FactorTreeNode));
  (*FTNode)->siblings[1] = dominatorFTNode;

#ifdef PTL
  /* MUX-related operation */
  if (operator == BDS_BDD_MUX && control != NULL) {
      controlFTNode = ALLOC(FactorTreeNode, 1); DBG_MEMSET (controlFTNode, 0, sizeof (FactorTreeNode));
      (*FTNode)->siblings[2] = controlFTNode;
  }
#endif

  /* Update this fnode */
  BDS_WriteFnode(*FTNode,p,operator,BDS_FTREE_INTERNAL);

  /* Decompose dp */
  result = BDS_DDDecomposeAll(bddmgr, decomposed, option,&decomposedFTNode); 
  if (result == 0) { return 0;}

#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_DDDecompose", 3);
#endif

  /* Decompose dm */
  result = BDS_DDDecomposeAll(bddmgr, dominator, option,&dominatorFTNode); 
  if (result == 0) { return 0;}

#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_DDDecompose", 4);
#endif

#ifdef PTL
  /* Decompose control */
  if (operator == BDS_BDD_MUX && control != NULL) {
      result = BDS_DDDecomposeAll(bddmgr, control, option,&controlFTNode); 
      if (result == 0) { return 0;}
  }
#endif

#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_DDDecompose", 5);
#endif

  return(1);

} /* end of BDS_DDDecompose */

/**Function********************************************************************

  Synopsis    [Program for BDD Decomposition-Based BDS]

  Description [An ADD is decomposed into *dominator and *decomposed at the best level,
	        the factoring tree is built along with the decomposition. 
		A queue is used to line up the BDD waited to be decomposed.

		All the possible cuts on the 0-1 ADD are tried, and both Boolean
		divisor and subtractor is evaluated based on that cut. The BEST
		Boolean divisor or subtractor is found by storing all the COST
		of each cut. The BEST has the LEAST cost. 

		Special case, like 0-, 1- and x-dominators are detected separately.

		return 1 success; 0 otherwise]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static
int
BDS_DDDecomposeAll(
  DdManager *bddmgr,
  bddPool *topSop,
  bdsOptions *option,
  FactorTreeNode **FTNode)
{
  int result, operator, sl;
  bddPool *decomposed = NULL, *dominator = NULL,*control = NULL; 
  DdNode *decompBdd;
  FactorTreeNode *decomposedFTNode, *dominatorFTNode, *controlFTNode, *currentFTNode, *FTNodeInCache;
  bddPool *toBeDecomposed, *topElementQ, *tmpElementQ;
  SQueue *decompositionQ;
  st_table *sopFTNodeTbl;
  int p; /* If decompBdd is complementary or not after reordering */

  /* Initialize a Q for bddPool* */
  decompositionQ = ALLOC(SQueue, 1);
  BDS_Q_Init(decompositionQ);
  BDS_EQ(decompositionQ, topSop);

  /* Initialize a hash table to store bddPool* vs FTNode* association */
  sopFTNodeTbl = st_init_table(st_ptrcmp,st_ptrhash);
  if (st_add_direct(sopFTNodeTbl, (char *)topSop, (char *)(*FTNode)) == ST_OUT_OF_MEM) {
      printf("Out of memory !");
      exit(2);
  }

  while ( (toBeDecomposed = BDS_DQ(decompositionQ)) != NULL ) {

      /* Build the BDD from SOP form */
      decompBdd = BDS_BuildDDFromBddPool(bddmgr, toBeDecomposed, 1);

      if (decompBdd == NULL) { return(0);}
      if (Cudd_IsConstant(decompBdd)) {
          fail("Can not decompose a constant node !!\n"); 
      }

      /* Obtain the currentFTNode from association tbl before it is freed */
      if (!st_delete(sopFTNodeTbl, (char **) &toBeDecomposed, (char **) &currentFTNode)) {
	  return(0);
      }

      /* Free the elementQ immediately. */
      BDS_FreeBddPool(bddmgr, toBeDecomposed); 

      /* When the BDD has one variable, reach a terminal node, go to next Q */
      /* otherwise leave currentFTNode uninitialized */
      result = BDS_CheckTerminal(bddmgr,decompBdd,currentFTNode);
      if (result != 0) continue;

      /* Decompose the BDD */
      result = BDS_BddDecomposeMain(bddmgr,decompBdd,option,&decomposed,&dominator,&control,&operator,&p);
      if (result == 0) { return 0;}

      /* Enqueue decomposed and dominator SOPs and build the factor tree*/
      decomposedFTNode = ALLOC(FactorTreeNode, 1); 
      DBG_MEMSET (decomposedFTNode, 0, sizeof (FactorTreeNode));
      currentFTNode->siblings[0] = decomposedFTNode;
      BDS_EQ(decompositionQ, decomposed);
      if (st_add_direct(sopFTNodeTbl, (char *)decomposed, (char *) decomposedFTNode) == ST_OUT_OF_MEM) {
	  return(0);
      }

      dominatorFTNode = ALLOC(FactorTreeNode, 1); 
      DBG_MEMSET (dominatorFTNode, 0, sizeof (FactorTreeNode));
      currentFTNode->siblings[1] = dominatorFTNode;
      BDS_EQ(decompositionQ, dominator);
      if (st_add_direct(sopFTNodeTbl, (char *)dominator, (char *) dominatorFTNode) == ST_OUT_OF_MEM) {
	  return(0);
      }

#ifdef PTL
      /* MUX-related operation */
      if (operator == BDS_BDD_MUX && control != NULL) {
          controlFTNode = ALLOC(FactorTreeNode, 1); 
          DBG_MEMSET (controlFTNode, 0, sizeof (FactorTreeNode));
	  currentFTNode->siblings[2] = controlFTNode;
	  BDS_EQ(decompositionQ, control);
          if (st_add_direct(sopFTNodeTbl, (char *)control, (char *) controlFTNode) == ST_OUT_OF_MEM) {
	      return(0);
          }
      }
#endif

      /* Update this fnode */
      BDS_WriteFnode(currentFTNode,p,operator,BDS_FTREE_INTERNAL);

  } /* end of while loop */

  BDS_Q_Quit(decompositionQ);
  st_free_table(sopFTNodeTbl);

  return(1); 

} /* end of BDS_DDDecomposeAll*/

/**Function********************************************************************

  Synopsis    [Main Bdd decomposition program for BDS]

  Description [result = 1, success; 0 otherwise]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static 
int 
BDS_BddDecomposeMain(
  DdManager *bddmgr,
  DdNode *bdd,
  bdsOptions *option,
  bddPool **decomposed,
  bddPool **dominator,
  bddPool **control,
  int *operator,
  int *polarity)
{
  int i, decompType, result, level;
  int decompLevel;
  int *support;
  int bddSize, numVar;
  edgeMark **edgeProperty = NULL;
  DdNode *add;
  DdNode *N; /* Regular version of bdd */

  bddStats *reorderMonitor;	         /* JS */
  if (BDS_CheckTimeout (option))     /* JS */
  {
    fprintf(stderr, "\nDecomposition timeout\n");
    fflush(stderr);
    fflush(stdout);
    exit(1);
  }
  
  /* Always pass the regular version of bdd for decomposition */
  N = Cudd_Regular(bdd);

  /* Avoid huge bdd */
  bddSize = Cudd_DagSize(N);
  numVar = Cudd_SupportSize(bddmgr, N);
  if (bddSize > numVar * 50) {
      option->reordering = CUDD_REORDER_SIFT;
      /* Record the reorder change if required JS */
      if (option->verb > BDS_VERBOSE_MOST) {
          reorderMonitor = ALLOC(bddStats, 1);
	      BDS_UpdateBddMonitor(bddmgr,reorderMonitor,bdd,NULL,NULL,NULL,0, 1.0, 0);
      }
      BDS_Reorder(bddmgr, NULL, option);
      if (option->verb > BDS_VERBOSE_MOST) {
          BDS_UpdateBddMonitor(bddmgr,reorderMonitor,NULL,bdd,NULL,NULL,0, 1.0, 0);
          BDS_Stats(BDS_REORDER,reorderMonitor,0,NULL,NULL,NULL,NULL);
          FREE(reorderMonitor);
      }
  }
  if (option->methods & BDS_CTRL_SID) {          /* JS */
      /* Transform incoming BDD into 0-1 ADD */
      add = Cudd_BddToAdd(bddmgr, N);
      Cudd_Ref(add);

      /* Record polarity after reordering */
      *polarity = Cudd_IsComplement(bdd) ? 0 : 1;

      support = BDS_SupportArray(bddmgr, N, &level);
      if (support == NULL) return(0);
      if (level<0) fail ("Negative level");
      edgeProperty = BDS_MarkEdges(bddmgr, add, support, level);
      if (edgeProperty == NULL) return(0);

      result = BDS_SimpleDominator(bddmgr,N,add,support,level,&decompLevel,edgeProperty,
                                         decomposed,dominator,operator,option);
                                         /* Discards the ADD if successful JS */
      BDS_FreeEdgeProperty(support,edgeProperty,level);
      if (result == 1) return(1);

      Cudd_RecursiveDeref(bddmgr,add);
  }
  /* Until here, no simple decomposition is found on the BDD, reorder the 
     BDD before decomposition 
  */
  bddSize = Cudd_DagSize(N);
  numVar = Cudd_SupportSize(bddmgr, N);

  if ((option->effort > 3) && (bddSize * numVar < option->limitSize)) {
      option->reordering = CUDD_REORDER_EXACT;
  }
  else {
      option->reordering = CUDD_REORDER_SIFT;
  }
  /* Record the reorder change if required  JS */
  if (option->verb > BDS_VERBOSE_MOST) {
      reorderMonitor = ALLOC(bddStats, 1);
      BDS_UpdateBddMonitor(bddmgr,reorderMonitor,bdd,NULL,NULL,NULL,0, 1.0, 0);
  }
  BDS_Reorder(bddmgr, NULL, option);
  if (option->verb > BDS_VERBOSE_MOST) {
      BDS_UpdateBddMonitor(bddmgr,reorderMonitor,NULL,bdd,NULL,NULL,0, 1.0, 0);
      BDS_Stats(BDS_REORDER,reorderMonitor,0,NULL,NULL,NULL,NULL);
      FREE(reorderMonitor);
  }

  *polarity = Cudd_IsComplement(bdd) ? 0 : 1;

  add = Cudd_BddToAdd(bddmgr, N);
  Cudd_Ref(add);

  support = BDS_SupportArray(bddmgr, N, &level);
  if (support == NULL) return(0);
      if (level<0) fail ("Negative level");
  edgeProperty = BDS_MarkEdges(bddmgr, add, support, level);
  if (edgeProperty == NULL) return(0);

  if (option->methods & BDS_CTRL_SID) {          /* JS */
      result = BDS_SimpleDominator(bddmgr,N,add,support,level,&decompLevel,edgeProperty,
                                         decomposed,dominator,operator,option);
      if (result == 1) {
          BDS_FreeEdgeProperty(support,edgeProperty,level);
          return(1);
      }
  }
  /* Search for branch decomposition. Perform branch decomposition based on 
     branches found on the BDD.
  */
  if (option->methods & BDS_CTRL_BHC) {          /* JS */
      result = BDS_BHardCore(bddmgr,N,add,option,decomposed,dominator,control,operator,support,level,edgeProperty);
      if (result == 1) {
          BDS_FreeEdgeProperty(support,edgeProperty,level);
          return(1);
      }
  }

  /* If the BDD is unbalanced, decompose it using topVar */
  result = BDS_TestBalance(bddmgr, N);
  if (result == 1) { /* Decompose with respect to topVar */
      Cudd_RecursiveDeref(bddmgr,add);
      if (option->methods & BDS_CTRL_CFT) {          /* JS */
          result = BDS_CofactorOnTop(bddmgr,bdd,option,decomposed,dominator,control,operator,polarity);
          BDS_FreeEdgeProperty(support,edgeProperty,level);
          return(result);
      } else {
          BDS_FreeEdgeProperty(support,edgeProperty,level);
          return 0;
      }
  }

  /* Perform decompositions based on generalized dominators */
  if (option->methods & BDS_CTRL_GED) {          /* JS */
      result = BDS_FindBestDecomp(bddmgr,N,add,support,option,decomposed,dominator,operator,level,edgeProperty); 
      BDS_FreeEdgeProperty(support,edgeProperty,level);
      if (result == 0) return(0);               /* means internal failure JS */
  } else {
      result = BDS_BDD_NO_DECOMP;               /* emulate unsuccessful decomp JS */
  }

  if (result == BDS_BDD_NO_DECOMP || result == BDS_BDD_NO_MINIMI || result == BDS_DECOMP_DISQUALIFY) {

      support = BDS_SupportArray(bddmgr, N, &level);
      if (support == NULL) return(0);
      if (level<0) fail ("Negative level");
      edgeProperty = BDS_MarkEdges(bddmgr, N, support, level);
      if (edgeProperty == NULL) return(0);

      if (option->methods & BDS_CTRL_XHC) {          /* JS */
         result = BDS_XHardCore(bddmgr,N,option,decomposed,dominator,operator,support,level,edgeProperty);
         if (result == 1) {
            BDS_FreeEdgeProperty(support,edgeProperty,level);
	        return(1);
         }
      }

      BDS_FreeEdgeProperty(support,edgeProperty,level);

      /* Temply put it here, cyang 1/17/99 */
      /* debugDump(bddmgr,bdd,0,option); */

      /* No efficient decomposition are detected, take cofactor with respect to topVar */
      if (option->methods & BDS_CTRL_CFT) {          /* JS */
          result = BDS_CofactorOnTop(bddmgr,bdd,option,decomposed,dominator,control,operator,polarity);
          if (result == 0) return(0);
      } else {
          return 0;                                 /* emulate no decomp JS */
      }
  }

  return(1);

} /* end of BDS_BddDecomposeMain */ 

/**Function********************************************************************

  Synopsis    [Find the best decomposition of a BDD using generalized dominators]

  Description [A BDD is decomposed at all possible level. Four cases may happen,
	BDS_BDD_DECOMP_NONE : when no edge Sigma;
	BDS_BDD_DECOMP_CONJ : when edge Sigma_0;
	BDS_BDD_DECOMP_DISJ : when edge Sigma_1;
	BDS_BDD_DECOMP_BOTH : when edges Sigma_1 AND Sigma_0.

	The function takes BDDs in the form of bddPool*, return decomposed
	and dominator also in the form of bddPool*. If the best decomposition cost
	is larger than a predefined theshold, the decomposition is disqualified.
	Under this case, the BDD is most likely a hard core. If a good decomposition
	can not be found on this bdd, the corresponding add is freed.

	Only regular version of bdd is decomposed. add is freed at the end.]

  SideEffects []

  SeeAlso     []

  LastDate    [2/23/99]

******************************************************************************/
static
int
BDS_FindBestDecomp(
  DdManager *bddmgr,
  DdNode *N,		/* The BDD to be decomposed. */
  DdNode *add,
  int *support,
  bdsOptions *option,
  bddPool **decomposed, /* Q or R */
  bddPool **dominator,  /* D */
  int *operator,
  int level,
  edgeMark **edgeProperty)
{
  int decompType, result, i;
  int numSigmaZero, numSigmaOne;
  decompCost *bestDecompCost;
  bddStats *bddMonitor;
  int decompLevel;
  bddPool *decomposedSOP, *dominatorSOP;
  DdNode *decomposedBdd, *dominatorBdd;

  /* Prepare to record decomposition information */
  if (option->verb > BDS_VERBOSE_MOST) {
      bddMonitor = ALLOC(bddStats, 1);
      BDS_UpdateBddMonitor(bddmgr,bddMonitor,N,NULL,NULL,NULL,0,0,0);
  }

  bestDecompCost = BDS_InitBestCost();
  numSigmaZero = 0; numSigmaOne = 0;

  /* Foreach level find out the decomposition and update the best cost. */
  for (i=0; i < level; i++) {

     /* perform decomposition at level i */
     if (edgeProperty[i]->decompType == BDS_BDD_DECOMP_NONE) continue;
     if ( (edgeProperty[i]->decompType == BDS_BDD_DECOMP_CONJ) ||
		(edgeProperty[i]->decompType == BDS_BDD_DECOMP_BOTH) ) {

	/* See if need to do conjunctive decomposition by checking number of
	   Sigma_0. Since the number of Sigma_0 on level(i+1) >= level(i), if
	   |Sigma_0| on level(i+1) = |Sigma_0| on level(i), it means the level(i+1)
	   has the same set of Sigma_0 with level(i), therefore no need to do
	   decomposition on this level. 
  	*/
 	if (numSigmaZero < edgeProperty[i]->numSigmaZero) { /* Do conjunctive decomposition. */
	    numSigmaZero = edgeProperty[i]->numSigmaZero;
	    result = BDS_PerformDecomp(bddmgr, N, add, edgeProperty[i], &bestDecompCost,
  					option,	BDS_BDD_DECOMP_CONJ, i);
	    if (result == 0) return(0);
	}
     }
     if ( (edgeProperty[i]->decompType == BDS_BDD_DECOMP_DISJ) ||
		(edgeProperty[i]->decompType == BDS_BDD_DECOMP_BOTH) ) {

	if (numSigmaOne < edgeProperty[i]->numSigmaOne) { /* Do disjunctive decomposition */
	    numSigmaOne = edgeProperty[i]->numSigmaOne;
	    result = BDS_PerformDecomp(bddmgr, N, add, edgeProperty[i], &bestDecompCost,
  					option,	BDS_BDD_DECOMP_DISJ, i);
	    if (result == 0) return(0);
   	}
     }  
  } 

  /* Disqualify this decomposition if the cost is too high */
  if (bestDecompCost->cost > BDS_DISQUALIFY_THRESHOLD &&
		bestDecompCost->dominator != NULL) { /* Not valid decomposition */
      if (option->verb > BDS_VERBOSE_MOST) {
          BDS_UpdateBddMonitor(bddmgr,bddMonitor,NULL,NULL,NULL,NULL,0,bestDecompCost->cost,0);
	  BDS_Stats(BDS_BDD_DECOMP,bddMonitor,0,NULL,NULL,NULL,NULL);
	  FREE(bddMonitor);
      }

      Cudd_RecursiveDeref(bddmgr,add);
      BDS_FreeBddPool(bddmgr, bestDecompCost->dominator);
      FREE(bestDecompCost);
      return(BDS_DECOMP_DISQUALIFY);
  }

  /* There is some possibility that no Sigam_0 and Sigma_1 exist on this BDD 
  */
  if ((numSigmaZero == 0) && (numSigmaOne == 0)) {
      FREE(bestDecompCost);
      Cudd_RecursiveDeref(bddmgr, add);
      return(BDS_BDD_NO_DECOMP);
  }

  if (bestDecompCost->dominator != NULL) { /* Construct return parameters */

      *dominator = bestDecompCost->dominator;
      *operator = bestDecompCost->operator;

      dominatorBdd = BDS_BuildDDFromBddPool(bddmgr, *dominator, 0);
      if (dominatorBdd == NULL) return(0);
      decomposedBdd = BDS_BddMinimization(bddmgr, N, dominatorBdd, 
						bestDecompCost->decompType, option); 
      if (decomposedBdd == NULL) return(0);
      if (decomposedBdd == N) { /* The same BDD is returned. No minimization was done */

          BDS_FreeBddPool(bddmgr, bestDecompCost->dominator);
          FREE(bestDecompCost);
          Cudd_RecursiveDeref(bddmgr, dominatorBdd);
          Cudd_RecursiveDeref(bddmgr, decomposedBdd);
          Cudd_RecursiveDeref(bddmgr, add);
          return(BDS_BDD_NO_MINIMI);
      }

      *decomposed = BDS_DumpBddPool(bddmgr, decomposedBdd);
      if (decomposed == NULL) return(0);

      /* Record the decomposition if required */
      if (option->verb > BDS_VERBOSE_MOST) {
	  BDS_UpdateBddMonitor(bddmgr,bddMonitor,NULL,decomposedBdd,dominatorBdd,NULL,
		bestDecompCost->level,bestDecompCost->cost,bestDecompCost->operator);
	  BDS_Stats(BDS_G_DOMI,NULL,0,NULL,NULL,NULL,NULL);
	  BDS_Stats(BDS_BDD_DECOMP,bddMonitor,0,NULL,NULL,NULL,NULL);
      }
      else if (option->verb > BDS_VERBOSE_LESS) {
	  BDS_Stats(BDS_G_DOMI,NULL,0,NULL,NULL,NULL,NULL);
      }

      Cudd_RecursiveDeref(bddmgr, dominatorBdd);
      Cudd_RecursiveDeref(bddmgr, decomposedBdd);
      Cudd_RecursiveDeref(bddmgr, N);
      Cudd_RecursiveDeref(bddmgr, add);

      FREE(bestDecompCost);
      if (option->verb > BDS_VERBOSE_MOST) FREE(bddMonitor);

      return(1);
  }
  else {
      Cudd_RecursiveDeref(bddmgr, add);
      FREE(bestDecompCost);
      if (option->verb > BDS_VERBOSE_MOST) FREE(bddMonitor);

      return(BDS_BDD_NO_MINIMI);
  }

} /* end of BDS_FindBestDecomp */

/**Function********************************************************************

  Synopsis    [Check if a BDD is a constant or has single variable]

  Description [If the bdd is a terminal, the corresponding fnode is
	updated inside this function. 
	Return 1, the bdd is a termianl; 0 not]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static 
int 
BDS_CheckTerminal(
  DdManager *bddmgr,
  DdNode *tBdd,
  FactorTreeNode *fnode)
{

  if (Cudd_SupportSize(bddmgr, tBdd) == 1 || Cudd_IsConstant(tBdd)) {
      /* Build the element in a fnode */
      if (cuddT(Cudd_Regular(tBdd)) == Cudd_ReadOne(bddmgr) && 
		cuddE(Cudd_Regular(tBdd)) == Cudd_ReadLogicZero(bddmgr)) {
          BDS_WriteFnode (fnode,1,0,Cudd_NodeReadIndex(tBdd));
      }
      else if (cuddT(Cudd_Regular(tBdd)) == Cudd_ReadLogicZero(bddmgr) && 
		cuddE(Cudd_Regular(tBdd)) == Cudd_ReadOne(bddmgr)) {
          BDS_WriteFnode(fnode,0,0,Cudd_NodeReadIndex(tBdd));
      }
      else if (tBdd == Cudd_ReadOne(bddmgr)) { /* Constant 1. */
          BDS_WriteFnode(fnode,0,0,BDS_FTREE_ONE);
      }
      else if (tBdd == Cudd_ReadLogicZero(bddmgr)) { /* Constant 0 */
          BDS_WriteFnode(fnode,0,0,BDS_FTREE_ZERO);
      }
      else {
          fail("Unknown BDD value.");
      }
      if (Cudd_IsComplement(tBdd) && !Cudd_IsConstant(tBdd)) {
          BDS_WriteFnode(fnode, 1- fnode->polarity, fnode->operator, fnode->value);
      }

      fnode->siblings[0] = NULL; fnode->siblings[1] = NULL; fnode->siblings[2] = NULL;

      Cudd_RecursiveDeref(bddmgr, tBdd);
      return(1);

  } /* end of when BDD is a constant or single variable */

  return(0);

} /* end of BDS_CheckTerminal */

/**Function********************************************************************

  Synopsis    [Check if there are any special properties on the BDD. Currently,
	0-dominator, 1-dominator and x-dominator are checked. Once there is a
	dominator existing on this BDD or 0-1 ADD, the decomposition is algebraic,
	perform the decomposition immediately.
    Discards the ADD if successful JS]

  Description [return 1, find a simple dominator; 0 otherwise]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static 
int
BDS_SimpleDominator(
  DdManager   	*bddmgr, 
  DdNode 	*N, 		/* always in regular, therefore no xor */
  DdNode 	*add,
  int		*support,
  int		level,		/* The total level of incoming BDD */
  int		*decompLevel,	/* The level where decomposition is taken */
  edgeMark	**edgeProperty,
  bddPool 	**decomposed,   /* Q or R */
  bddPool	**dominator,    /* D */
  int 		*operator,
  bdsOptions *option)
{
  DdNode *specialDominator = NULL, *f; /* for special check */
  bddPool *decomposedSOP, *dominatorSOP;
  int i, result, specialOperator;
  bddStats *bddMonitor;

  /* Search for 0, 1 and x-dominators 
  */
  if (specialDominator = (DdNode *) BDS_SpecialCheck(bddmgr,N,add,support,level,
				decompLevel,edgeProperty,&specialOperator,option)) {
      switch(specialOperator) {
      case BDS_BDD_OR:
	   f = BDS_ORDecompose(bddmgr,N,specialDominator,option);
	   if (f == 0) fail("Serious error in BDS_FindBestDecomp()");
	   break;
      case BDS_BDD_AND:
	   f = BDS_ANDDecompose(bddmgr,N,specialDominator,option);
	   if (f == 0) fail("Serious error in BDS_FindBestDecomp()");
	   break;
      case BDS_BDD_XNOR:
	   f = BDS_XDecompose(bddmgr,N,specialDominator,&specialOperator,option);
	   if (f == 0) fail("Serious error in BDS_FindBestDecomp()");
	   break;
      default:
	   fail("Unknown decomposition type in BDS_FindBestDecomp()");
      }

      decomposedSOP = BDS_DumpBddPool(bddmgr, f);
      if (decomposedSOP == NULL) fail("Error in BDS_FindBestDecomp()");
      *decomposed = decomposedSOP;

      dominatorSOP = BDS_DumpBddPool(bddmgr, specialDominator);
      if (dominatorSOP == NULL) fail("Error in BDS_FindBestDecomp()");
      *dominator = dominatorSOP;

      *operator = specialOperator;

      /* Record the decomposition statistical information if required */
      if (option->verb > BDS_VERBOSE_MOST) {
          bddMonitor = ALLOC(bddStats, 1);
          BDS_UpdateBddMonitor(bddmgr,bddMonitor,N,NULL,NULL,NULL,0,0,0);
	  BDS_UpdateBddMonitor(bddmgr,bddMonitor,NULL,f,specialDominator,NULL,
						*decompLevel,0,specialOperator);
	  BDS_Stats(BDS_BDD_DECOMP,bddMonitor,0,NULL,NULL,NULL,NULL);
	  FREE(bddMonitor);
      }

      Cudd_RecursiveDeref(bddmgr, f);
      Cudd_RecursiveDeref(bddmgr, specialDominator);
      Cudd_RecursiveDeref(bddmgr, N);
      Cudd_RecursiveDeref(bddmgr, add);

      return(1);
  } 

  return(0);

} /* end of BDS_SimpleDominator */

/**Function********************************************************************

  Synopsis    [Initialize a cost monitor for generalized dominator-based decomp]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/
static
decompCost *
BDS_InitBestCost(
   )
{
  decompCost *cost;

  cost = ALLOC(decompCost, 1);
  cost->dominator = NULL;
  cost->decomposed = NULL;
  cost->cost = BDS_HUGE_NUMBER;
  cost->numQR = 0;
  cost->numD = 0;

  return(cost);

} /* end of BDS_InitBestCost */

/**Function********************************************************************

  Synopsis    [Initialize a cost monitor for X decomp]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/
static
XdecompCost *
BDS_InitXdecompCost(
  )
{
  XdecompCost *cost;

  cost = ALLOC(XdecompCost, 1);
  cost->cost = BDS_HUGE_NUMBER;
  cost->f = NULL;
  cost->g = NULL;
  cost->h = NULL;

  return(cost);

} /* end of BDS_InitXdecompCost */

/**Function********************************************************************

  Synopsis    [Perform a decomposition on a level]

  Description [The decomposition result is compared with bestDecompCost. The best
	decomposition's level, bddPool are stored in bestDecompCost.

	result = 1, success; 0 otherwise]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static
int
BDS_PerformDecomp(
  DdManager *bddmgr,
  DdNode *bdd,			/* BDD to be decomposed. */
  DdNode *add,			/* BDD in 0-1 ADD form. */
  edgeMark *edgeTbl, 		/* edges on this level are stored in this hash table. */
  decompCost **bestDecompCost,
  bdsOptions *option,
  int decompType,		/* decompType is AND or OR */
  int level)
{
  bddPool *dominatorSOP;
  DdNode *dominatorBdd, *decomposedBdd;
  int result, i;
  decompCost *thisCost;

  dominatorSOP = BDS_DumpBddPoolAtLevel(bddmgr, add, NULL, edgeTbl, decompType);
  if (dominatorSOP == NULL) return(0);
  dominatorBdd = BDS_BuildDDFromBddPool(bddmgr, dominatorSOP, 0);
  if (dominatorBdd == NULL) return(0);

  /* Do BDD minimization. */
  decomposedBdd = BDS_BddMinimization(bddmgr, bdd, dominatorBdd, decompType, option); 

  if (decomposedBdd == NULL) {
      Cudd_RecursiveDeref(bddmgr, dominatorBdd);
      BDS_FreeBddPool(bddmgr, dominatorSOP);
      return(0);
  }
  if (decomposedBdd == bdd) { /* No minimization is done, don't update cost. */
      Cudd_RecursiveDeref(bddmgr, dominatorBdd);
      Cudd_RecursiveDeref(bddmgr, decomposedBdd);
      BDS_FreeBddPool(bddmgr, dominatorSOP);
      return(BDS_BDD_NO_MINIMI);
  }

  thisCost = BDS_CalculateDecompCost(bddmgr, bdd, decomposedBdd, dominatorBdd, option);

  /* Update the best decomp found so far. A balanced decomposition is prefered. */
  if ((thisCost->cost < (*bestDecompCost)->cost) ||
	((thisCost->cost == (*bestDecompCost)->cost) && 
	(thisCost->numQR * thisCost->numD > (*bestDecompCost)->numQR * (*bestDecompCost)->numD))) {

      if ((*bestDecompCost)->dominator != NULL) { /* avoid the 1st iteration error. */
	  BDS_FreeBddPool(bddmgr, (*bestDecompCost)->dominator);
      }
      switch(decompType) {
      case BDS_BDD_DECOMP_CONJ:
          (*bestDecompCost)->operator = BDS_BDD_AND;
          break;
      case BDS_BDD_DECOMP_DISJ:
	  (*bestDecompCost)->operator = BDS_BDD_OR;
	  break;
      default: 
	  fail("Unknown decomposition type in BDS_PerformDecomp");
      }
      (*bestDecompCost)->cost = thisCost->cost;
      (*bestDecompCost)->dominator = dominatorSOP;
      (*bestDecompCost)->level = level;
      (*bestDecompCost)->numQR = thisCost->numQR;
      (*bestDecompCost)->numD = thisCost->numD;
      (*bestDecompCost)->decompType = decompType;
  }
  else {
      BDS_FreeBddPool(bddmgr, dominatorSOP);
  }
  Cudd_RecursiveDeref(bddmgr, decomposedBdd);
  Cudd_RecursiveDeref(bddmgr, dominatorBdd);	
  FREE(thisCost);
 
  return(1);

} /* end of BDS_PerformDecomp */

/**Function********************************************************************

  Synopsis    [Calculate the cost of a decomposition]

  Description [Return cost struct]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static
decompCost *
BDS_CalculateDecompCost(
  DdManager *bddmgr,
  DdNode *oriBdd,
  DdNode *decBdd,
  DdNode *domBdd,
  bdsOptions *option)
{
  DdNode *oriSupp, *decSupp, *domSupp, *commSupp;
  int result, oriSize, decSize, domSize, oriSuppSize, commSuppSize;
  decompCost *thisCost;

  thisCost = ALLOC(decompCost, 1);

  oriSize = Cudd_DagSize(oriBdd);
  decSize = Cudd_DagSize(decBdd);
  domSize = Cudd_DagSize(domBdd);

  result = Cudd_ClassifySupport(bddmgr, decBdd, domBdd, &commSupp, &decSupp, &domSupp);
  if (result == 0) return(NULL);
  Cudd_Ref(commSupp); Cudd_Ref(decSupp); Cudd_Ref(domSupp);

  commSuppSize = Cudd_SupportSize(bddmgr, commSupp);
  oriSuppSize = Cudd_SupportSize(bddmgr, decSupp) +
		Cudd_SupportSize(bddmgr, domSupp) + commSuppSize;

  Cudd_RecursiveDeref(bddmgr, commSupp);
  Cudd_RecursiveDeref(bddmgr, decSupp);
  Cudd_RecursiveDeref(bddmgr, domSupp);

  thisCost->cost = ALPHA * (decSize + domSize) / oriSize + (1 - ALPHA) * commSuppSize / oriSuppSize;  
  thisCost->numQR = decSize;
  thisCost->numD = domSize;

  return(thisCost);

} /* end of BDS_CalculateDecompCost */

/**Function********************************************************************

  Synopsis    [Calculate the cost of a X decomposition]

  Description [Return cost]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static
float
BDS_CalXdecompCost(
  DdManager *bddmgr,
  DdNode *bdd,		/* Original bdd */
  int base,		/* Negative edges on bdd */
  DdNode *f,
  int nef,		/* ne on f */
  DdNode *g,
  int neg,		/* ne on g */
  DdNode *h,
  int decompType)
{
  DdNode *fSupp, *gSupp, *hSupp, *commSupp, *fgCommSupp, *fhCommSupp;
  int result;
  int oriSize, oriSuppSize, fSize, gSize, hSize, fSuppSize, gSuppSize, hSuppSize, commSuppSize;
  int fgCommSuppSize, fhCommSuppSize;
  float cost;

  fSize = (float) Cudd_DagSize(f);

  if (f != NULL) gSize = Cudd_DagSize(g);
  if (h != NULL) hSize = Cudd_DagSize(h);

  oriSize = Cudd_DagSize(bdd);
  oriSuppSize = Cudd_SupportSize(bddmgr, bdd);

  switch(decompType) {
      case BDS_BDD_DECOMP_X:
          result = Cudd_ClassifySupport(bddmgr, f, g, &commSupp, &fSupp, &gSupp);
          if (result == 0) exit(2);

          Cudd_Ref(commSupp); Cudd_Ref(fSupp); Cudd_Ref(gSupp);
  	  commSuppSize = Cudd_SupportSize(bddmgr, commSupp);

  	  Cudd_RecursiveDeref(bddmgr, commSupp);
  	  Cudd_RecursiveDeref(bddmgr, fSupp);
  	  Cudd_RecursiveDeref(bddmgr, gSupp);

	  cost = BETA * (fSize + gSize) / oriSize 
		      + (1.0 - BETA) * commSuppSize / oriSuppSize 
		      + (float) (nef + neg) / base;  
	  break;

      case BDS_BDD_DECOMP_B:
          result = Cudd_ClassifySupport(bddmgr, f, g, &fgCommSupp, &fSupp, &gSupp);
          if (result == 0) exit(2);

          Cudd_Ref(fgCommSupp); Cudd_Ref(fSupp); Cudd_Ref(gSupp);
  	  fgCommSuppSize = Cudd_SupportSize(bddmgr, fgCommSupp);

  	  Cudd_RecursiveDeref(bddmgr, fgCommSupp);
  	  Cudd_RecursiveDeref(bddmgr, fSupp);
  	  Cudd_RecursiveDeref(bddmgr, gSupp);

          result = Cudd_ClassifySupport(bddmgr, f, h, &fhCommSupp, &fSupp, &hSupp);
          if (result == 0) exit(2);

          Cudd_Ref(fhCommSupp); Cudd_Ref(fSupp); Cudd_Ref(hSupp);
  	  fhCommSuppSize = Cudd_SupportSize(bddmgr, fhCommSupp);

  	  Cudd_RecursiveDeref(bddmgr, fhCommSupp);
  	  Cudd_RecursiveDeref(bddmgr, fSupp);
  	  Cudd_RecursiveDeref(bddmgr, hSupp);

	  cost = BETA * (fSize + gSize + hSize) / oriSize 
	              + (1.0 - BETA) * (fgCommSuppSize + fhCommSuppSize) / oriSuppSize;
	  break;

      default:
	  printf("Unknown decomposition type !\n");
	  exit(2);
  }

  return(cost);

} /* end of BDS_CalXdecompCost */

/**Function********************************************************************

  Synopsis    [Check if there are any special properties on the BDD. Currently,
	0-dominator, 1-dominator and x-dominator are checked.] 

  Description [Dominator closest to median is returned]

  SideEffects []

  SeeAlso     []

  LastDate    [2/23/99]

*****************************************************************************/
static 
DdNode *
BDS_SpecialCheck(
  DdManager   	*bddmgr, 
  DdNode 	*bdd,           /* always in regular, therefore no xor */
  DdNode 	*add,
  int		*support,
  int		level,		/* The total level of incoming BDD */
  int		*decompLevel,	/* The level where decomposition is taken */
  edgeMark	**edgeProperty,
  int 		*operator,
  bdsOptions *option)
{
  DdNode *dominator;

  /* 0-dominator */
  if (dominator = (DdNode *) BDS_SearchZeroD(bddmgr,support,level,decompLevel,edgeProperty)) {
      *operator = BDS_BDD_OR;
      if (option->verb > BDS_VERBOSE_LESS) {
          BDS_Stats(BDS_ZERO_DOMI,NULL,0,NULL,NULL,NULL,NULL);
      }
      return(dominator);
  }
  /* 1-diminator */
  if (dominator = (DdNode *) BDS_SearchOneD(bddmgr,support,level,decompLevel,edgeProperty)) {
      *operator = BDS_BDD_AND;
      if (option->verb > BDS_VERBOSE_LESS) {
          BDS_Stats(BDS_ONE_DOMI,NULL,0,NULL,NULL,NULL,NULL);
      }
      return(dominator);
  }
  /* x-dominator */
  if (dominator = (DdNode *) BDS_SearchXD(bddmgr, bdd, decompLevel)) {
      *operator = BDS_BDD_XNOR;
      if (option->verb > BDS_VERBOSE_LESS) {
          BDS_Stats(BDS_X_DOMI,NULL,0,NULL,NULL,NULL,NULL);
      }
      return(dominator);
  }
  return(NULL);

} /* end of BDS_SpecialCheck */

/**Function********************************************************************

  Synopsis    [Detect if there is a 0-dominator]

  Description [0-dominator closest to median is returned]

  SideEffects []

  SeeAlso     []

  LastDate    [2/23/99]

*****************************************************************************/
static 
DdNode *
BDS_SearchZeroD(
  DdManager *bddmgr, 
  int *support,
  int level,
  int *decompLevel,
  edgeMark **edgeProperty) 
{
  int i, median, position, distance, best = 0xFFFF;
  DdNode *dominator = NULL, *addNode;

  if (edgeProperty == NULL) return(NULL);

  median = level / 2;

  for (i=0; i < level; i++) {
      if ((edgeProperty[i]->decompType == BDS_BDD_DECOMP_DISJ) &&
				(st_count(edgeProperty[i]->table) == 1)) {

	  /* A 0-dominator is found */
	  if (edgeProperty[i]->bottom == NULL) { return(NULL); }
	  distance = abs(i - median);
	  if (distance < best) {
              position = i;
	      best = distance;
	      if (best == 0) break;
	  }
	  else { /* Stop searching */
              break;
	  }
      }
  }

  if (best != 0xFFFF) { /* Found a 0-dominator */
      addNode = edgeProperty[position]->bottom;
      dominator = Cudd_addBddPattern(bddmgr, addNode);
      Cudd_Ref(dominator);
      *decompLevel = position;
  }

  return(dominator);

} /* end of BDS_SearchZeroD */

/**Function********************************************************************

  Synopsis    [Detect if there is a 1-dominator]

  Description [1-dominator closest to median is returned]

  SideEffects []

  SeeAlso     []

  LastDate    [2/23/99]

*****************************************************************************/
static 
DdNode *
BDS_SearchOneD(
  DdManager *bddmgr, 
  int *support,
  int level,
  int *decompLevel,
  edgeMark **edgeProperty) 
{
  int i, median, position, distance, best = 0xFFFF;
  DdNode *dominator = NULL, *addNode;

  if (edgeProperty == NULL) return(NULL);

  median = level / 2;

  for (i=0; i < level; i++) {
      if ((edgeProperty[i]->decompType == BDS_BDD_DECOMP_CONJ) &&
				(st_count(edgeProperty[i]->table) == 1)) {

	  /* A 1-dominator is found */
	  if (edgeProperty[i]->bottom == NULL) { return(NULL); }
	  distance = abs(i - median);
	  if (distance < best) {
              position = i;
	      best = distance;
	      if (best == 0) break;
	  }
	  else { /* Stop searching */
              break;
	  }
      }
  }

  if (best != 0xFFFF) { /* Found a 0-dominator */
      addNode = edgeProperty[position]->bottom;
      dominator = Cudd_addBddPattern(bddmgr, addNode);
      Cudd_Ref(dominator);
      *decompLevel = position;
  }

  return(dominator);

} /* end of BDS_SearchOneD */

/**Function********************************************************************

  Synopsis    [Detect if there are any x-dominator on this BDD]

  Description []

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static 
DdNode *
BDS_SearchXD(
  DdManager *bddmgr, 
  DdNode *bdd,
  int *decompLevel) 
{
  int *support, level, i, result;
  int median, position, distance, best = 0xFFFF;
  edgeMark **edgeProperty = NULL;
  DdNode *xDominator = NULL;

  support = BDS_SupportArray(bddmgr, bdd, &level);
  if (support == NULL) { return(0);}

  /* Init. a hash table for each level */
  if (level<0) fail ("Negative level");
  edgeProperty = BDS_MarkEdges(bddmgr, bdd, support, level);
  if (edgeProperty == NULL) return(0);

  median = level / 2;

  for (i=0; i < level; i++) {
      if ((edgeProperty[i]->decompType == BDS_BDD_DECOMP_NONE) &&
				(st_count(edgeProperty[i]->table) == 1)) {

	  /* A x-dominator is found */
	  if (edgeProperty[i]->bottom == NULL) { return(NULL); }
	  distance = abs(i - median);
	  if (distance < best) {
              position = i;
	      best = distance;
	      if (best == 0) break;
	  }
	  else { /* Stop searching */
              break;
	  }
      }
  }

  if (best != 0xFFFF) {
      xDominator = edgeProperty[position]->bottom;
      Cudd_Ref(xDominator);
      *decompLevel = position;
  }

  for (i = 0; i < level; i++) {
      st_free_table(edgeProperty[i]->table);
      FREE(edgeProperty[i]);
  }
  FREE(edgeProperty);
  FREE(support);

  return(xDominator);

} /* end of BDS_SearchXD */

/**Function********************************************************************

  Synopsis    [Reorder the BDD to find out x-dominator]

  Description [Return 1 if found a x-dominator, return 0 otherwise. The incoming
	bdd is freed if found a x-dominator]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static 
int
BDS_OneMoreTry(
  DdManager *bddmgr, 
  DdNode *bdd, 
  bdsOptions *option,
  bddPool **decomposed,
  bddPool **dominator,
  int *operator,
  int *polarity)
{
  int bddSize, numVar, result, decompLevel;
  DdNode *xDominator, *f, *N;

  N = Cudd_Regular(bdd);    

  /* Since there are hard cores in this BDD, whenever the size is less than
     the predefined value, apply exact. 
  */
  bddSize = Cudd_DagSize(bdd);
  numVar = Cudd_SupportSize(bddmgr, bdd);
  if (bddSize * numVar < option->limitSize) {
      option->reordering = CUDD_REORDER_EXACT;
      BDS_Reorder(bddmgr, NULL, option);

      *polarity = Cudd_IsComplement(bdd) ? 0 : 1;
      N = Cudd_Regular(bdd);

      if (xDominator = (DdNode *) BDS_SearchXD(bddmgr, N, &decompLevel)) {
      	  f = BDS_XDecompose(bddmgr,N,xDominator,operator,option);
          if (f == NULL) {
	      Cudd_RecursiveDeref(bddmgr,xDominator);
	      return(0);
	  }

	  /* Build return values, operator is updated inside BDS_XDecompose() */
	  *decomposed = BDS_DumpBddPool(bddmgr, xDominator);
	  *dominator = BDS_DumpBddPool(bddmgr, f);

	  Cudd_RecursiveDeref(bddmgr,xDominator);
	  Cudd_RecursiveDeref(bddmgr,f);
	  Cudd_RecursiveDeref(bddmgr,bdd);

	  return(1);
      }
      else {
	  return(0); 
      }
  }
  return(0);

} /* end of BDS_OneMoreTry */

/**Function********************************************************************

  Synopsis    [Perform OR decomposition on a BDD.]

  Description [Since F = f + g, g is given, then f = F/g']

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static 
DdNode *
BDS_ORDecompose(
  DdManager *bddmgr,
  DdNode *bdd,
  DdNode *cBdd,
  bdsOptions *option)
{
  DdNode *retBdd;

  retBdd = BDS_BddMinimization(bddmgr,bdd,cBdd,BDS_BDD_DECOMP_DISJ,option);
  if (retBdd == NULL) return(NULL);

  return(retBdd);

} /* end of BDS_ORDecompose */

/**Function********************************************************************

  Synopsis    [Perform AND decomposition on a BDD.]

  Description [Since F = fg, g is given, then f = F/g]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static 
DdNode *
BDS_ANDDecompose(
  DdManager *bddmgr,
  DdNode *bdd,
  DdNode *cBdd,
  bdsOptions *option)
{
  DdNode *retBdd;

  retBdd = BDS_BddMinimization(bddmgr,bdd,cBdd,BDS_BDD_DECOMP_CONJ,option);
  if (retBdd == NULL) return(NULL);

  return(retBdd);

} /* end of BDS_ANDDecompose */

/**Function********************************************************************

  Synopsis    [Perform XOR decomposition on a BDD.]

  Description [F = f xnor d; d is given, then f = (Fd)/d]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static 
DdNode *
BDS_XDecompose(
  DdManager *bddmgr, 
  DdNode *bdd, 
  DdNode *xDominator,
  int *operator,
  bdsOptions *option)
{
  DdNode *F, *d, *u, *tmp, *retBdd;

  d = xDominator;

  F = Cudd_Regular(bdd);
  if (Cudd_IsComplement(bdd)) {
      *operator = BDS_BDD_XOR;
  }
  else {
      *operator = BDS_BDD_XNOR;
  }

  tmp = Cudd_bddAnd(bddmgr, F, d);
  if (tmp == NULL) return(NULL);
  Cudd_Ref(tmp);

  retBdd = BDS_BddMinimization(bddmgr,tmp,d,BDS_BDD_DECOMP_CONJ,option);
  if (retBdd == NULL) return(NULL);
  Cudd_RecursiveDeref(bddmgr, tmp);
  
  return(retBdd);

} /* BDS_XDecompose */

/**Function********************************************************************

  Synopsis    [Perform XOR decomposition]

  Description [Usually this type of BDDs have a lot of negative edges. It's unlikely
	to have non-algebraic decomposition on these BDDs. All algebraic cases have
	been found in simple dominators. Two types action can be taken here, 
	F = f @ g or F = f ^ g(based on BDD), and F = fg + f'h(based on ADD). 
	These types of decompositions are good candidates for PTL.

	Return 1, successful; 0 otherwise]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static
int
BDS_XHardCore(
  DdManager *bddmgr,
  DdNode *bdd,		/* The BDD to be decomposed. */
  bdsOptions *option,
  bddPool **f, 
  bddPool **g, 
  int *operator,
  int *support,
  int level,
  edgeMark **edgeProperty)
{
  int result, i, negativeEdge;
  XdecompCost *bestDecomp;
  bddStats *bddMonitor;

  negativeEdge = BDS_GetNegative(bddmgr,bdd,0);
  if (negativeEdge == 0) return(0);

  /* Foreach level find out all XOR decomp */
  bestDecomp = BDS_InitXdecompCost();
  for (i=0; i < level; i++) {
      result = BDS_XDecompOnBdd(bddmgr,bdd,negativeEdge,support,i,edgeProperty[i],bestDecomp,option);
  }
  if (bestDecomp->f != NULL && bestDecomp->g != NULL) {

      *f = BDS_DumpBddPool(bddmgr, bestDecomp->f);
      *g = BDS_DumpBddPool(bddmgr, bestDecomp->g);
      *operator = BDS_BDD_XNOR;

      /* Record the decomposition statistical information if required */
      if (option->verb > BDS_VERBOSE_MOST) {
          bddMonitor = ALLOC(bddStats, 1);
          BDS_UpdateBddMonitor(bddmgr,bddMonitor,bdd,bestDecomp->f,bestDecomp->g,NULL,0,
							bestDecomp->cost,*operator);
          BDS_Stats(BDS_XHARDCORE,bddMonitor,0,NULL,NULL,NULL,NULL);
          FREE(bddMonitor);
      }
      else if (option->verb > BDS_VERBOSE_LESS) {
      	  BDS_Stats(BDS_XHARDCORE,NULL,0,NULL,NULL,NULL,NULL);
      }

      /* Free related BDDs */
      Cudd_RecursiveDeref(bddmgr,bestDecomp->f);
      Cudd_RecursiveDeref(bddmgr,bestDecomp->g);
      Cudd_RecursiveDeref(bddmgr,bdd);
      FREE(bestDecomp);
      return(1);
  }
  else {
      FREE(bestDecomp);
      return(0);
  }
  
} /* end of BDS_XHardCore */

/**Function********************************************************************

  Synopsis    [Perform branch decomposition]

  Description [Usually this type of BDDs are good candidates for F = fg + f'h;
	       The decomposition can be more effciently implemented as PTL.
	       Both static CMOS and PTL style are supported in this function.
	       In static CMOS case, fg is returned as decomposed, f'h is returned
	       as dominators. The coverage between g and h is tested to save
	       one more literal. In PTL case, f is returned as control, g is returned
	       as decomposed and h is returned as dominator.

               Return 1, successful; 0 otherwise]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static
int
BDS_BHardCore(
  DdManager *bddmgr,
  DdNode *bdd,		/* The BDD to be decomposed. */
  DdNode *add,
  bdsOptions *option,
  bddPool **decomposed,	/* g or fg */
  bddPool **dominator, 	/* h or f'h */
  bddPool **control,	/* f or useless in static CMOS */
  int *operator,	/* BDS_BDD_MUX or BDS_BDD_OR */
  int *support,
  int level,
  edgeMark **edgeProperty)
{
  int result, i, negativeEdge, lowest = 0;
  DdNode *f, *g, *h;
  #ifndef PTL
  DdNode *fg, *fh, *cover;
  #endif
  DdNode *gBranch, *hBranch;
  XdecompCost *bestDecomp;
  bddStats *bddMonitor;

  /* Find the lowest two-branch level. Start from level 1, instead of 0 */
  for (i = 1; i < level; i++) {
      if (st_count(edgeProperty[i]->table) == 2 && edgeProperty[i]->numSigmaOne == 0
						&& edgeProperty[i]->numSigmaZero == 0) {
	  lowest = i;
      }
  }
  if (lowest == 0) return(0);

  /* Generate f */
  f = BDS_BranchCandidateB(bddmgr, add, edgeProperty[lowest], &gBranch, &hBranch);

  /* Calculate g and h */
  g = Cudd_addBddPattern(bddmgr, gBranch);
  if (g == NULL) return(0);
  Cudd_Ref(g);
  Cudd_RecursiveDeref(bddmgr,gBranch);

  h = Cudd_addBddPattern(bddmgr, hBranch);
  if (h == NULL) return(0);
  Cudd_Ref(h);
  Cudd_RecursiveDeref(bddmgr, hBranch);

#ifdef PTL

  /* Dump f, g and h in bddPool */
  *decomposed = BDS_DumpBddPool(bddmgr, g);
  *dominator = BDS_DumpBddPool(bddmgr, h);
  *control = BDS_DumpBddPool(bddmgr, f);
  *operator = BDS_BDD_MUX;

#else

  /* Find out if g covers h, or reverse */
  cover = Cudd_bddAnd(bddmgr, g, h);
    if (cover == NULL) return(0);
      Cudd_Ref(cover);

      if (cover == g) { /* g < h */
	  fg = g;
	  Cudd_Ref(fg);
	  fh = Cudd_bddAnd(bddmgr, Cudd_Not(f), h);
	  if (fh != NULL) Cudd_Ref(fh);
      }
      else if (cover == h) { /* g > h*/
	  fg = Cudd_bddAnd(bddmgr, f, g);
	  if (fg != NULL) Cudd_Ref(fg);
	  fh = h;
	  Cudd_Ref(fh);
      }
      else { /* g and h don't cover each other */
	  fg = Cudd_bddAnd(bddmgr, f, g);
	  if (fg != NULL) Cudd_Ref(fg);
	  fh = Cudd_bddAnd(bddmgr, Cudd_Not(f), h);
	  if (fh != NULL) Cudd_Ref(fh);
      }
      *decomposed = BDS_DumpBddPool(bddmgr, fg);
      *dominator = BDS_DumpBddPool(bddmgr, fh);
      *operator = BDS_BDD_OR;

#endif

  /* Record the decomposition if required */
  if (option->verb > BDS_VERBOSE_MOST) {
      bddMonitor = ALLOC(bddStats, 1);
      BDS_UpdateBddMonitor(bddmgr,bddMonitor,bdd,g,h,f,0,0,0);
      BDS_Stats(BDS_BHARDCORE,bddMonitor,0,NULL,NULL,NULL,NULL);
      FREE(bddMonitor);
  }
  else if (option->verb > BDS_VERBOSE_LESS) {
      BDS_Stats(BDS_BHARDCORE,NULL,0,NULL,NULL,NULL,NULL);
  } 

  Cudd_RecursiveDeref(bddmgr, g);
  Cudd_RecursiveDeref(bddmgr, h);
  Cudd_RecursiveDeref(bddmgr, f);

#ifndef PTL
  Cudd_RecursiveDeref(bddmgr, cover);
  Cudd_RecursiveDeref(bddmgr, fg);
  Cudd_RecursiveDeref(bddmgr, fh);
#endif
  
  Cudd_RecursiveDeref(bddmgr, bdd);
  Cudd_RecursiveDeref(bddmgr, add);

  return(1);

} /* end of BDS_BHardCore */

/**Function********************************************************************

  Synopsis    [Find candidate for 2-branch decomposition on BDD(with ne case)]

  Description [BDD is returned if found; return NULL otherwise]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static
DdNode *
BDS_BranchCandidateX(
  DdManager *bddmgr, 
  DdNode *bdd,
  edgeMark *levelTable) 
{
  DdNode *xcand;
  bddPool *pool;

  /* A possible branch-candidate is found */
  pool = BDS_DumpBddPoolAtLevel(bddmgr, bdd, NULL, levelTable, BDS_BDD_DECOMP_X);
  xcand = BDS_BuildDDFromBddPool(bddmgr, pool, 0);
  BDS_FreeBddPool(bddmgr, pool);

  if (Cudd_IsConstant(xcand)) 
      return(NULL);
  else
      return(xcand);

} /* end of BDS_BranchCandidateX */

/**Function********************************************************************

  Synopsis    [Find candidate for 2-branch decomposition on BDD(without ne case)]

  Description [BDD is returned if found; return NULL otherwise]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static
DdNode *
BDS_BranchCandidateB(
  DdManager *bddmgr, 
  DdNode *add,
  edgeMark *levelTable,
  DdNode **gBranch,
  DdNode **hBranch) 
{
  DdNode *bcand, *oneNode, *f;
  bddPool *pool;
  st_table *oneMark;
  int result;
  st_generator *gen;

  oneMark = st_init_table(st_ptrcmp,st_ptrhash);

  oneNode = BDS_FindOneMark(bddmgr,add,levelTable);
  if (st_add_direct(oneMark, (char *) oneNode, NULL) == ST_OUT_OF_MEM) {
      printf("Out of memory !");
      exit(2);
  }
  
  gen = st_init_gen(levelTable->table);
  while (st_gen(gen, (char **) &f, NULL)) {
      if (f != oneNode) {
	  Cudd_Ref(f);
	  *hBranch = f;
          break;
      }
  }
  Cudd_Ref(oneNode);
  *gBranch = oneNode;
  st_free_gen(gen);

  pool = BDS_DumpBddPoolAtLevel(bddmgr, add, oneMark, levelTable, BDS_BDD_DECOMP_B);
  bcand = BDS_BuildDDFromBddPool(bddmgr, pool, 0);

  BDS_FreeBddPool(bddmgr, pool);
  st_free_table(oneMark);

  if (Cudd_IsConstant(bcand)) 
      return(NULL);
  else
      return(bcand);

} /* end of BDS_BranchCandidateB */

/**Function********************************************************************

  Synopsis    [Perform Branch decomposition on BDD. The decomposer f is obtained top down.]

  Description [F = fg + f'h is seeking; Return 1 if successful; 0 otherwise]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static
int
BDS_BranchDecompTopDown(
  DdManager *bddmgr,
  DdNode *bdd,
  edgeMark *levelTable)
{
  DdNode *f, *g, *h;
  DdNode *cover;

  f = BDS_BranchCandidateX(bddmgr, bdd, levelTable);
  if (f == NULL) return(0);

  g = Cudd_bddConstrain(bddmgr, bdd, f);
  if (g != NULL) Cudd_Ref(g);

  h = Cudd_bddConstrain(bddmgr, bdd, Cudd_Not(f));
  if (h != NULL) Cudd_Ref(h);

  cover = Cudd_bddAnd(bddmgr, g, h);
  if (cover != NULL) Cudd_Ref(cover);

  Cudd_RecursiveDeref(bddmgr, g);
  Cudd_RecursiveDeref(bddmgr, h);
  Cudd_RecursiveDeref(bddmgr, f);
  Cudd_RecursiveDeref(bddmgr, cover);

  return(0);

} /* end of BDS_BranchDecompTopDown */
 
/**Function********************************************************************

  Synopsis    [Perform Branch decomposition on BDD. The decomposer f is obtained bottom up.]

  Description [F = fg + f'h is seeking; Return 1 if successful; 0 otherwise]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static
int
BDS_BranchDecompBottomUp(
  DdManager *bddmgr,
  DdNode *bdd,
  int *support,
  int level,
  edgeMark *levelTable,
  XdecompCost *bestDecomp,
  bdsOptions *option)
{
  DdNode *f, *g, *h;
  DdNode *cover;
  float cost;
  st_generator *gen;
  int result, base, nef, neg, retVal = 0;

  gen = st_init_gen(levelTable->table);
  while (st_gen(gen, (char **) &f, NULL)) {

#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_BranchDecompBottomUp", 1);
#endif

      if (Cudd_Regular(f)->ref == 1) continue; 
      if (support[Cudd_Regular(f)->index] != level + 1) continue;
      Cudd_Ref(f);

       g = Cudd_bddConstrain(bddmgr, bdd, f);
       if (g != NULL) Cudd_Ref(g);

       h = Cudd_bddConstrain(bddmgr, bdd, Cudd_Not(f));
       if (h != NULL) Cudd_Ref(h);

       cost = BDS_CalXdecompCost(bddmgr,bdd,0,f,0,g,0,h,BDS_BDD_DECOMP_B);

#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_BranchDecompBottomUp", 1);
#endif

       if (cost > bestDecomp->cost) {
           Cudd_RecursiveDeref(bddmgr,g);
           Cudd_RecursiveDeref(bddmgr,h);
           Cudd_RecursiveDeref(bddmgr,f);
       }
       else {
           if (bestDecomp->cost < BDS_HUGE_NUMBER) {
               Cudd_RecursiveDeref(bddmgr,bestDecomp->g);
#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_BranchDecompBottomUp", 1);
#endif
               Cudd_RecursiveDeref(bddmgr,bestDecomp->h);
#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_BranchDecompBottomUp", 1);
#endif
               Cudd_RecursiveDeref(bddmgr,bestDecomp->f);

#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_BranchDecompBottomUp", 1);
#endif

           }
           bestDecomp->cost = cost;
           bestDecomp->operator = BDS_BDD_DECOMP_B;
           bestDecomp->f = f;
           bestDecomp->g = g;       
           bestDecomp->h = h;

           retVal = 1;
       }
  }
  st_free_gen(gen);
  
  return(retVal);

} /* end of BDS_BranchDecompBottomUp */

/**Function********************************************************************

  Synopsis    [Perform X decomposition on BDD]

  Description [A X candidate, f, is found on each level. Two types of decompositions
	are performed. F = f@g; and F = fh + f'i; 
	Return 1 if successful; 0 otherwise]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static
int
BDS_XDecompOnBdd(
  DdManager *bddmgr,
  DdNode *bdd,
  int base,		/* # of negative edges on the BDD */
  int *support,
  int level,
  edgeMark *levelTable,
  XdecompCost *bestDecomp,
  bdsOptions *option)
{
  DdNode *f, *g;
  float cost;
  st_generator *gen;
  int result, nef, neg, retVal = 0;

  bddStats *evalMonitor;	/* JS */

  gen = st_init_gen(levelTable->table);
  while (st_gen(gen, (char **) &f, NULL)) {

       if (support[Cudd_Regular(f)->index] != level + 1) continue;
       Cudd_Ref(f);
       nef = BDS_GetNegative(bddmgr,f,1);

       g = Cudd_bddXnor(bddmgr, f, bdd);
       if (g != NULL) Cudd_Ref(g);
       neg = BDS_GetNegative(bddmgr,g,1);

       if (nef + neg >= base) { /* Disqualify this decomp */
           Cudd_RecursiveDeref(bddmgr,g);
           Cudd_RecursiveDeref(bddmgr,f);
       } 
       else {                   /* reject a decomp that can cause loops JS */
           evalMonitor = ALLOC(bddStats, 1);
	       BDS_UpdateBddMonitor(bddmgr,evalMonitor,bdd,g,NULL,NULL,0, 1.0, 0);
           if (evalMonitor->dpVar == evalMonitor->inVar 
            && evalMonitor->dpSize - evalMonitor->inSize >= option->xdelta) {
               if (option->verb > BDS_VERBOSE_MOST) {   /* report rejection */
                   BDS_Stats(BDS_HACKOUT,evalMonitor,0,NULL,NULL,NULL,option);
               }
               Cudd_RecursiveDeref(bddmgr,g);
               Cudd_RecursiveDeref(bddmgr,f);
           }
	       else { /* Compare the decomp with the previous best one */
    		   cost = BDS_CalXdecompCost(bddmgr,bdd,base,f,nef,g,neg,NULL,BDS_BDD_DECOMP_X);
    	  	   if (cost >= bestDecomp->cost) {
    		       Cudd_RecursiveDeref(bddmgr,g);
    		       Cudd_RecursiveDeref(bddmgr,f);
    		   }
    		   else {
    		       if (bestDecomp->cost < BDS_HUGE_NUMBER) {
    			   Cudd_RecursiveDeref(bddmgr,bestDecomp->f);
    			   Cudd_RecursiveDeref(bddmgr,bestDecomp->g);
    		       }
    		       bestDecomp->cost = cost;
    		       bestDecomp->operator = BDS_BDD_XNOR;
    		       bestDecomp->f = f;
    		       bestDecomp->g = g;	
    		       bestDecomp->h = NULL;

    		       retVal = 1;
    		   }
	       }
           FREE(evalMonitor);
       }
  }
  st_free_gen(gen);
  
  return(retVal);

} /* end of BDS_XDecompOnBdd */

/**Function********************************************************************

  Synopsis    [Perform GX decomposition on BDD]

  Description [Decomposition F = f@g + h is seeking; f is known. 
	F = fg + f'g' + h; Ff = f(g+h); Ff' = f'(g'+h); Therefore,
	h = (Ff/f)(Ff'/f');

	g = (F/h')@f;

	Intermediate variables: i = Ff; j = Ff'; k = Ff/f; l = Ff'/f';
				m = F/h';

	Return 1 if successful; 0 otherwise]

  SideEffects []

  SeeAlso     []

*****************************************************************************/
static
int
BDS_GXDecompOnBdd(
  DdManager *bddmgr,
  DdNode *bdd,
  int *support,
  int level,
  edgeMark *levelTable,
  bdsOptions *option)
{
  DdNode *f, *g, *h;
  DdNode *i, *j, *k, *l, *m;
  DdNode *try1, *try2;
  st_generator *gen;

  gen = st_init_gen(levelTable->table);
  while (st_gen(gen, (char **) &f, NULL)) {
       if (Cudd_Regular(f)->ref == 1) continue; 
       if (support[Cudd_Regular(f)->index] != level + 1) continue;
       Cudd_Ref(f);

       i = Cudd_bddAnd(bddmgr, bdd, f);
       if (i != NULL) Cudd_Ref(i);

       j = Cudd_bddAnd(bddmgr, bdd, Cudd_Not(f));
       if (j != NULL) Cudd_Ref(j);

       k = Cudd_bddLICompaction(bddmgr, i, f);
       if (k != NULL) Cudd_Ref(k);

       l = Cudd_bddLICompaction(bddmgr, j, Cudd_Not(f));
       if (l != NULL) Cudd_Ref(l);

       h = Cudd_bddAnd(bddmgr, k, l);
       if (h != NULL) Cudd_Ref(h);

       m = Cudd_bddLICompaction(bddmgr, bdd, Cudd_Not(h));
       if (m != NULL) Cudd_Ref(m);

       g = Cudd_bddXnor(bddmgr, f, m);
       if (g != NULL) Cudd_Ref(g);

       /* Debug */
       try1 = Cudd_bddXnor(bddmgr, f, g);
       if (try1 != NULL) Cudd_Ref(try1);

       try2 = Cudd_bddOr(bddmgr, try1, h);
       if (try2 != NULL) Cudd_Ref(try2);

       Cudd_RecursiveDeref(bddmgr,f);
       Cudd_RecursiveDeref(bddmgr,g);
       Cudd_RecursiveDeref(bddmgr,h);
       Cudd_RecursiveDeref(bddmgr,i);
       Cudd_RecursiveDeref(bddmgr,j);
       Cudd_RecursiveDeref(bddmgr,k);
       Cudd_RecursiveDeref(bddmgr,l);
       Cudd_RecursiveDeref(bddmgr,m);
       Cudd_RecursiveDeref(bddmgr,try1);
       Cudd_RecursiveDeref(bddmgr,try2);
  }
  st_free_gen(gen);
  
  return(1);

} /* end of BDS_GXDecompOnBdd */

/**Function********************************************************************

  Synopsis    [Co-factor with respect to the top variable, this a basic MUX decomposition]

  Description [Decomposition F = fg + f'h; Both static CMOS and PTL style are
  supported in this function. Return 1 if successful; 0 otherwise]

  SideEffects []

  SeeAlso     []

  Last Date   [6/15/98]

*****************************************************************************/
static 
int 
BDS_CofactorOnTop(
  DdManager *bddmgr, 
  DdNode *bdd, 
  bdsOptions *option,
  bddPool **decomposed, 
  bddPool **dominator, 
  bddPool **control,
  int *operator,
  int *polarity)
{
  DdNode *N, *var;
  DdNode *g, *h;
  #ifndef PTL
  DdNode *fg, *fh, *cover;
  #endif
  int result;
  bddStats *bddMonitor;
  
  /* Always pass the regular version of bdd for decomposition */
  *polarity = Cudd_IsComplement(bdd) ? 0 : 1;
  N = Cudd_Regular(bdd);

  /* Find top variable */
  var = bddmgr->vars[N->index];
  cuddRef(var);

  /* Find co-factors */
  g = cuddT(N);
  cuddRef(g);
  h = cuddE(N);
  cuddRef(h);

#ifdef PTL

  /* Dump f, g and h into bddPool */
  *decomposed = BDS_DumpBddPool(bddmgr, g);
  *dominator = BDS_DumpBddPool(bddmgr, h);
  *control = BDS_DumpBddPool(bddmgr, var);
  *operator = BDS_BDD_MUX;

#else

  cover = Cudd_bddAnd(bddmgr, g, h);
  if (cover == NULL) return(0);
  cuddRef(cover);

  if (cover == g) { /* g < h */
      fg = g;
      cuddRef(fg);
      fh = Cudd_bddAnd(bddmgr, Cudd_Not(var), h);
      if (fh != NULL) cuddRef(fh);
  }
  else if (cover == h) { /* g > h */
      fg = Cudd_bddAnd(bddmgr, var, g);
      if (fg != NULL) cuddRef(fg);
      fh = h;
      cuddRef(fh);
  }
  else { /* f and g don't cover each other */
      fg = Cudd_bddAnd(bddmgr, var, g);
      if (fg != NULL) cuddRef(fg);
      fh = Cudd_bddAnd(bddmgr, Cudd_Not(var), h);
      if (fh != NULL) cuddRef(fh);
  }
  *decomposed = BDS_DumpBddPool(bddmgr, fg);
  *dominator = BDS_DumpBddPool(bddmgr, fh);
  *operator = BDS_BDD_OR;

#endif

  /* Record if required */
  if (option->verb > BDS_VERBOSE_MOST) {
      bddMonitor = ALLOC(bddStats, 1);
      BDS_UpdateBddMonitor(bddmgr,bddMonitor,N,g,h,var,0,0,0);
      BDS_Stats(BDS_COFACTOR,bddMonitor,0,NULL,NULL,NULL,NULL);
      FREE(bddMonitor);
  }
  else {
      BDS_Stats(BDS_COFACTOR,NULL,0,NULL,NULL,NULL,NULL);
  }

  Cudd_RecursiveDeref(bddmgr, g);
  Cudd_RecursiveDeref(bddmgr, h);
  Cudd_RecursiveDeref(bddmgr, var);

#ifndef PTL
  Cudd_RecursiveDeref(bddmgr, cover);
  Cudd_RecursiveDeref(bddmgr, fg);
  Cudd_RecursiveDeref(bddmgr, fh);
#endif

  Cudd_RecursiveDeref(bddmgr, bdd);

  return(1);

} /* end of BDS_CofactorOnTop */

/**Function********************************************************************

  Synopsis    [Find out the node which should be connected to 1]

  Description [This function is only used for branch decomposition]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static
DdNode *
BDS_FindOneMark(
  DdManager *bddmgr,
  DdNode *bdd,
  edgeMark *levelTable)
{
  DdNode *oneMark;
  st_table *visited;

  visited = st_init_table(st_ptrcmp, st_ptrhash);

  oneMark = BDS_FindOneMarkRecursive(bddmgr, bdd, visited, levelTable);

  st_free_table(visited);

  return(oneMark);

} /* end of BDS_FindOneMark */

/**Function********************************************************************

  Synopsis    [Recursive program for BDS_FindOneMark()]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/
static
DdNode *
BDS_FindOneMarkRecursive(
  DdManager *bddmgr,
  DdNode *bdd,
  st_table *visited,
  edgeMark *levelTable)
{
  DdNode *N, *oneMark;

  N = Cudd_Regular(bdd);

  if (st_is_member(levelTable->table, (char *) N) == 1) { 
      return(N);
  }

  if (st_is_member(visited, (char *) N) == 1) { 
      return(NULL);
  }

  oneMark = BDS_FindOneMarkRecursive(bddmgr, cuddT(N), visited, levelTable);
  if (oneMark != NULL) return(oneMark);

  oneMark = BDS_FindOneMarkRecursive(bddmgr, cuddE(N), visited, levelTable);
  if (oneMark != NULL) return(oneMark);

  if (st_add_direct(visited, (char *) N, NULL) == ST_OUT_OF_MEM) {
      printf("Out of memeory !");
      exit(2);
  }

} /* end of BDS_FindOneMarkRecursive */

