/*************************************************************************
	-----------------------------------------------------
           		Copyright 2002, 
	   University of Massachusetts, Amherst. 
	-----------------------------------------------------

This package was created at the University of Massachusetts, Amherst. 
The University of Massachusetts makes no warranty about the suitability 
of this software for any purpose. It is presented on an "as is" basis.

CONDITIONS:

The use of software for commercial purposes is strictly forbidden
without a written consent from its authors or the University of 
Massachusetts, Amherst.

**************************************************************************/
/**CFile***********************************************************************

  FileName    [minimize.c]

  PackageName [BDS]

  Synopsis    [BDD Decomposition-Based Logic Optimization Program]

  Description []

  SeeAlso     []

  Author      [Anda Congguang Yang]

  Copyright   [The file was created in Department of Electrical & Computer
	Engineering, University of Massachusetts Amherst. All source codes
	in BDS package are CONFIDENTIAL before a formal release is
	announced. The author reserve all right of the package.]

******************************************************************************/

#include "lopt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Performs safe minimization of a BDD.]

  Description [Return minimized BDD if successful, NULL otherwise]

  SideEffects []

  SeeAlso     []

******************************************************************************/
extern
DdNode *
BDS_BddMinimization(
  DdManager *bddmgr,
  DdNode *f,		/* f BDD */
  DdNode *c,		/* c BDD */
  int decompType,	/* t = R + c OR f = Q c. */
  bdsOptions *option)
{
  int result;
  DdNode *retBdd;

  if (decompType == BDS_BDD_DECOMP_CONJ) {
      if (option->miniMethod == BDS_BDD_LICOMPACTION) {
	  retBdd = Cudd_bddLICompaction(bddmgr, f, c);
      }
      else { 
	  retBdd = Cudd_bddRestrict(bddmgr, f, c);
      }
  }
  else if (decompType == BDS_BDD_DECOMP_DISJ) {
      if (option->miniMethod == BDS_BDD_LICOMPACTION) {
	  retBdd = Cudd_bddLICompaction(bddmgr, f, Cudd_Not(c));
      }
      else {
	  retBdd = Cudd_bddRestrict(bddmgr, f, Cudd_Not(c));
      }
  }
  else {
      fail("This type of decomposition is not supported now !");
  }

  if (retBdd == NULL) { return(NULL);}
  Cudd_Ref(retBdd);

#if DEBUG_ACTIVE(DEBUG_CUDD_CHECK)
  cuddCheck(bddmgr, "BDS_BddMinimization", 2);
#endif

#if DEBUG_ACTIVE(DEBUG_DECOMP_CHECK)
  result = bdsDecompCheck(bddmgr,f,decompType,retBdd,c);
  if (result == 0) fail("BDD minimization error!");
#endif

  return(retBdd);

} /* BDS_BddMinimization */
